#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-08-31 14:04:53
Last Modified by: Hanyu Wang
Last Modified time: 2023-08-31 14:24:19
"""

from .truth_table import *
from .sum_of_product import *
