#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-15 13:25:54
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-15 13:28:55
"""

from .diff_states import get_difference
from .verify_circuit import verify_circuit_and_count_cnot, simulate_circuit
